
/* $Id: main.c,v 1.7 2001/02/19 21:47:04 mpalczew Exp $ 
 *** finder applet
 *** Copyright (C) 2000,2001 Mike Palczewski mpalczew@u.washington.edu
 *** 
 *** This program is free software; you can redistribute it and/or
 *** modify it under the terms of the GNU General Public License
 *** as published by the Free Software Foundation; either version 2
 *** of the License, or (at your option) any later version.
 *** 
 *** This program is distributed in the hope that it will be useful,
 *** but WITHOUT ANY WARRANTY; without even the implied warranty of
 *** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *** GNU General Public License for more details.
 *** 
 *** You should have received a copy of the GNU General Public License
 *** along with this program; if not, write to the Free Software
 *** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 *** 02111-1307, USA.
 ***/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk/gdkx.h>
#include <applet-widget.h>
#include <gnome.h>
#include <config.h>
#include "gwmh.h"
#include "properties.h"
#include "conf.h"
#include "pixmaps.h"
#include "unknown.xpm"

typedef enum{ICONIFY,SHADE,STICK,CLOSE,KILL} winaction;

typedef struct {
	GwmhTask *gwmhtask;
	GtkWidget *icon;
	GtkWidget *item;
	GtkWidget *label;
	Pixmap pixmap;
	GdkPixbuf *cache;
} mytask;

GtkWidget *get_icon(mytask * task, gint size);
GdkPixbuf *get_x_icon(mytask * task, gint size);
GtkWidget *get_pixmap_from_pixbuf(GdkPixbuf * pixbuf);
mytask *find_task(GwmhTask * gwmh_task);
gboolean on_click(GtkWidget * applet, GdkEventButton * event,
		  gpointer user_data);
void position(GtkMenu * menu, gint * x, gint * y, gpointer user_data);
void add_menu_item(gchar*  text,gchar** xpm,winaction action);
gboolean on_item_click(GtkWidget *item, GdkEventButton *event,
		gpointer user_data);
void make_action_menu(mytask* task);
int on_menu(GtkMenuItem * menuitem, gpointer user_data);
void on_action_menu(GtkMenuItem * menuitem, gpointer user_data);
void init_task(GwmhTask * gwmh_task);
void doabout(AppletWidget * widget, gpointer data);
void restore_grabs (GtkWidget *w,gpointer data);
gint task_notifier(gpointer func_data, GwmhTask * gwmh_task,
		   GwmhTaskNotifyType ntype, GwmhTaskInfoMask imask);
void destroy_task(GwmhTask * gwmh_task);
void update_task(GwmhTask * gwmh_task);
GdkPixbuf* get_mini_icon(GwmhTask * task);
static gpointer get_typed_property_data (Display *xdisplay, Window xwindow,
		Atom property, Atom requested_type, gint *size_p,
		guint expected_format);
void restore_border(void);
Pixmap get_pixmap(mytask* task);
void update_task_icon(GwmhTask *gwmh_task);
void on_size(AppletWidget* applet,int size);

GtkWidget *applet;
GList *tasklist = NULL;
GtkWidget *menu;
GtkWidget *label;
GtkWidget *pfrm;
GtkWidget *frame;
GtkWidget *action_menu=NULL;
mytask *current = NULL;
gchar *normal_path;
mytask *action_task;
conf *config;

void on_size(AppletWidget* applet,int size) {
	gtk_widget_set_usize(GTK_WIDGET(applet),config->width,applet->size);
}


Pixmap get_pixmap(mytask* task) {
	XWMHints *wmhints;
        Pixmap pixmap;

        wmhints = XGetWMHints (GDK_DISPLAY (), task->gwmhtask->xwin);

        if (!wmhints)
                return 0;

        if (!(wmhints->flags & IconPixmapHint)) {
                XFree (wmhints);
                return 0;
        }

        pixmap = wmhints->icon_pixmap;

        XFree (wmhints);

        return pixmap;
}
void restore_grabs (GtkWidget *w,gpointer data) {
	/* Stolen from menu.c from the gnome panel */
	GdkCursor *cursor=gdk_cursor_new(GDK_ARROW);
	GTK_MENU_SHELL(menu)->have_xgrab=
		(gdk_pointer_grab (menu->window, TRUE,
			GDK_BUTTON_PRESS_MASK |
			GDK_BUTTON_RELEASE_MASK |
			GDK_ENTER_NOTIFY_MASK |
			GDK_LEAVE_NOTIFY_MASK,
			NULL, cursor, 0) == 0);
	gtk_grab_add(GTK_WIDGET(menu));
}

gboolean on_item_click(GtkWidget *item, GdkEventButton *event, 
		gpointer user_data) 
{
		
	if (event->button == 3) {
		make_action_menu((mytask*)user_data);	
		action_task=(mytask*)user_data;
		gtk_signal_connect(GTK_OBJECT(action_menu),"deactivate",
				GTK_SIGNAL_FUNC(restore_grabs),NULL);
		gtk_menu_popup(GTK_MENU(action_menu), NULL, NULL, NULL, NULL, 3,
			       event->time);
		return TRUE;
	} else
		return FALSE;	

}

void make_action_menu(mytask* task)
{
	GtkWidget* item;
	if (action_menu)
		gtk_widget_destroy(action_menu);
	action_menu = gtk_menu_new();
	gtk_widget_show(action_menu);

	if ( GWMH_TASK_ICONIFIED(task->gwmhtask))
		add_menu_item("Restore",restore_xpm,ICONIFY);
	else  add_menu_item("Iconfify",iconify_xpm,ICONIFY);
	
	if ( GWMH_TASK_SHADED (task->gwmhtask))
		add_menu_item("Unshade",unshade_xpm,SHADE);
	else add_menu_item("Shade",shade_xpm,SHADE);

	if ( GWMH_TASK_STICKY (task->gwmhtask))
		add_menu_item("Unstick",unstick_xpm,STICK);
	else add_menu_item("Stick",stick_xpm,STICK);

	add_menu_item("Close",close_xpm,CLOSE);
	
	item=gtk_menu_item_new();
	gtk_widget_show(item);
	gtk_menu_append(GTK_MENU(action_menu),item);
		
	add_menu_item("Kill", kill_xpm,KILL);	
}

void add_menu_item(gchar*  text,gchar** xpm,winaction action) {
	GtkWidget *label;
	GtkWidget *item;
	GtkWidget *gtkpixmap;
	GdkPixmap *pixmap;
	GdkBitmap *mask;	
	
	item=gtk_pixmap_menu_item_new();
	label=gtk_label_new(text);
	gtk_misc_set_alignment(GTK_MISC(label),0,.5);
	pixmap=gdk_pixmap_create_from_xpm_d(applet->window, &mask, NULL
				, xpm);
	gtkpixmap=gtk_pixmap_new(pixmap,mask);
	gtk_container_add(GTK_CONTAINER(item),label);
	gtk_pixmap_menu_item_set_pixmap(GTK_PIXMAP_MENU_ITEM(item),gtkpixmap);
	gtk_widget_show_all(item);
	gtk_menu_append(GTK_MENU(action_menu),item);
	gtk_signal_connect(GTK_OBJECT(item),"activate",
			GTK_SIGNAL_FUNC(on_action_menu),
			GINT_TO_POINTER(action));
}

void on_action_menu(GtkMenuItem * menuitem, gpointer user_data) {
	winaction action=GPOINTER_TO_INT(user_data);

	switch(action) {
		case ICONIFY:
			if (GWMH_TASK_ICONIFIED(action_task->gwmhtask)) {
				gwmh_task_deiconify(action_task->gwmhtask);
			} else
				gwmh_task_iconify(action_task->gwmhtask);
			break;
		case SHADE:
			if (GWMH_TASK_SHADED(action_task->gwmhtask)) 
				 gwmh_task_unset_gstate_flags(
						 action_task->gwmhtask,
						 GWMH_STATE_SHADED);
			else  
				gwmh_task_set_gstate_flags(
						 action_task->gwmhtask,
						 GWMH_STATE_SHADED);
			break;
		case STICK:
			if (GWMH_TASK_STICKY(action_task->gwmhtask))
				gwmh_task_unset_gstate_flags(
						 action_task->gwmhtask,
						 GWMH_STATE_STICKY);
			else  
				gwmh_task_set_gstate_flags(
						 action_task->gwmhtask,
						 GWMH_STATE_STICKY);
			break;
		case CLOSE:
			gwmh_task_close(action_task->gwmhtask);
			break;
		case KILL:
			gwmh_task_kill(action_task->gwmhtask);
			break;
		default:
	}	
}

void position(GtkMenu * menu, gint * x, gint * y, gpointer user_data)
{
	gint lx, ty, by;
	gint swidth, sheight;
	gint mwidth, mheight;

	gdk_window_get_position(GTK_WIDGET(applet)->window, &lx, &ty);
	swidth = gdk_screen_width();
	sheight = gdk_screen_height();
	mwidth = GTK_WIDGET(menu)->requisition.width;
	mheight = GTK_WIDGET(menu)->requisition.height;
	by = ty + applet->allocation.height;

	/* This could be more sophisticated
	 * The goal should be not to raise the menu over 
	 * the applet I should steal from gtk*/
	if (lx + mwidth > swidth)
		lx = swidth - mwidth;
	if (lx < 0)
		lx = 0;
	*x = lx;

	if (by + mheight < sheight) {
		*y = by;
		return;
	}

	if (ty - mheight > 0)
		*y = ty - mheight;
	else
		*y = 0;
}

void restore_border() {
	gtk_frame_set_shadow_type(GTK_FRAME(frame),config->shadow_type);
}

gboolean
on_click(GtkWidget * applet, GdkEventButton * event, gpointer user_data)
{
	if (event->button == 1 ) {
		if (config->shadow_type != 0) 
			gtk_frame_set_shadow_type(GTK_FRAME(frame),
					config->shadow_type-1);	
		else 
			gtk_frame_set_shadow_type(GTK_FRAME(frame),GTK_SHADOW_OUT);
		gtk_signal_connect(GTK_OBJECT(menu),"deactivate",
				GTK_SIGNAL_FUNC(restore_border),NULL);
		
		gtk_menu_popup(GTK_MENU(menu), NULL, NULL, position, 
				NULL, 1,event->time);
		return TRUE;
	} else
		return FALSE;
}

GdkPixbuf*
get_mini_icon(GwmhTask * task)
{
	GdkGC *gc;
	int x, y, b, width, height, depth;
	guint32 *atomdata;
	Window root;
	GdkImage *image;
	GdkPixmap *pixmap;
	GdkBitmap *mask;
	GdkPixbuf *pixbuf;
	gint size;
	static gulong KWM_WIN_ICON = 0;
	Display *xdisplay;
	guchar *data;
	
	if (!KWM_WIN_ICON) {
		KWM_WIN_ICON = gdk_atom_intern ("KWM_WIN_ICON", FALSE);
	}

	xdisplay = GDK_WINDOW_XDISPLAY (task->gdkwindow);
	
	atomdata = get_typed_property_data (xdisplay,
					    task->xwin,
					    KWM_WIN_ICON,
					    KWM_WIN_ICON,
					    &size,
					    32);
	if (!atomdata)
		return FALSE;
	
	if (!atomdata[0])
		return FALSE;

	/* Get icon size and depth */
	XGetGeometry (xdisplay, (Drawable)atomdata[0], &root, &x, &y,
		      &width, &height, &b, &depth);
	
	/* Create a new GdkPixmap and copy the mini icon pixmap to it */
	pixmap = gdk_pixmap_new (NULL, width, height, depth);
	gc = gdk_gc_new (pixmap);
	XCopyArea (GDK_WINDOW_XDISPLAY (pixmap), atomdata[0], GDK_WINDOW_XWINDOW (pixmap),
		   GDK_GC_XGC (gc), 0, 0, width, height, 0, 0);
	gdk_gc_destroy (gc);
	
	pixbuf = gdk_pixbuf_get_from_drawable (NULL,
					       pixmap,
					       gtk_widget_get_colormap (applet),
					       0, 0,
					       0, 0,
					       width, height);
	gdk_pixmap_unref (pixmap);
	
	if (atomdata[1]) {
		mask = gdk_pixmap_new (NULL, width, height, depth);
		gc = gdk_gc_new (mask);
		gdk_gc_set_background (gc, &applet->style->black);
		gdk_gc_set_foreground (gc, &applet->style->white);
		XCopyPlane (GDK_DISPLAY (), atomdata[1], GDK_WINDOW_XWINDOW (mask),
			    GDK_GC_XGC (gc), 0, 0, width, height, 0, 0, 1);
		gdk_gc_unref (gc);
		
		image = gdk_image_get (mask, 0, 0, width, height);
		g_return_val_if_fail (image != NULL, FALSE);
		
		pixbuf = gdk_pixbuf_add_alpha (pixbuf, FALSE, 0, 0, 0);
		
		data = gdk_pixbuf_get_pixels (pixbuf);
		for (y = 0; y < gdk_pixbuf_get_height (pixbuf); y++) 
			for (x = 0; x < gdk_pixbuf_get_width (pixbuf); x++) 
			{
				data += 3;
				*data++ = gdk_image_get_pixel (image, x, y) == 0 ? 0 : 255;
			}
		
		gdk_pixmap_unref (mask);
	}
	
	return pixbuf;	
}

GtkWidget *get_icon(mytask * task, gint size)
{
	GdkPixbuf *tmp, *scaled;
	GtkWidget *ret;

	tmp = get_x_icon(task, size);
	if (tmp == NULL ) tmp=get_mini_icon(task->gwmhtask);
	if (tmp == NULL) 
		tmp = gdk_pixbuf_new_from_xpm_data(unknown_xpm);
	task->cache=tmp;
	scaled = gdk_pixbuf_scale_simple(tmp, size, size, GDK_INTERP_HYPER);
	tmp = scaled;
	ret = get_pixmap_from_pixbuf(tmp);
	gdk_pixbuf_unref(tmp);
	return ret;
}

GtkWidget *get_pixmap_from_pixbuf(GdkPixbuf * pixbuf)
{
	GdkPixmap *pixmap;
	GdkBitmap *bitmap;
	GtkWidget *ret;

	if (pixbuf == NULL)
		return NULL;
	gdk_pixbuf_render_pixmap_and_mask(pixbuf, &pixmap, &bitmap, 100);
	ret = gtk_pixmap_new(pixmap, bitmap);
	gdk_pixmap_unref(pixmap);
	if (bitmap != NULL) gdk_bitmap_unref(bitmap);
	gtk_widget_show(ret);
	return ret;
}

GdkPixbuf *get_x_icon(mytask * task, gint size)
{
	XWMHints *wmhints;
	GdkPixmap *pixmap, *mask;
	GdkPixbuf *pixbuf, *scaled;
	GdkImage *image;
	GdkGC *gc;
	Window root;
	gint x, y;
	guint width, height;
	guint border_width;
	guint depth;
	guchar *data;

	wmhints = XGetWMHints(GDK_DISPLAY(), task->gwmhtask->xwin);

	if (!wmhints) {
		return NULL;
	}

	if (!(wmhints->flags & IconPixmapHint)) {
		XFree(wmhints);
		return NULL;
	}

	XGetGeometry(GDK_DISPLAY(), wmhints->icon_pixmap, &root,
		     &x, &y, &width, &height, &border_width, &depth);
	if (width > 65535 || height > 65535) {
		XFree(wmhints);
		return NULL;
	}

	pixmap = gdk_pixmap_new(applet->window, width, height, -1);

	gc = gdk_gc_new(pixmap);
	if (depth == 1) {
		gdk_gc_set_background(gc, &applet->style->white);
		gdk_gc_set_foreground(gc, &applet->style->black);
		XCopyPlane(GDK_DISPLAY(), wmhints->icon_pixmap,
			   GDK_WINDOW_XWINDOW(pixmap), GDK_GC_XGC(gc), 0,
			   0, width, height, 0, 0, 1);
	} else {
		XCopyArea(GDK_DISPLAY(), wmhints->icon_pixmap,
			  GDK_WINDOW_XWINDOW(pixmap), GDK_GC_XGC(gc), 0, 0,
			  width, height, 0, 0);
	}

	pixbuf = gdk_pixbuf_get_from_drawable(NULL, pixmap,
					      gtk_widget_get_colormap
					      (applet), 0, 0, 0, 0, width,
					      height);
	gdk_gc_destroy(gc);

	if (depth == 1) {
		pixbuf = gdk_pixbuf_add_alpha(pixbuf, FALSE, 0, 0, 0);
	}

	if (wmhints->flags & IconMaskHint) {
		mask = gdk_pixmap_new(NULL, width, height, depth);
		gc = gdk_gc_new(mask);
		gdk_gc_set_background(gc, &applet->style->black);
		gdk_gc_set_foreground(gc, &applet->style->white);
		XCopyPlane(GDK_DISPLAY(), wmhints->icon_mask,
			   GDK_WINDOW_XWINDOW(mask), GDK_GC_XGC(gc), 0, 0,
			   width, height, 0, 0, 1);
		gdk_gc_unref(gc);

		image = gdk_image_get(mask, 0, 0, width, height);
		if (image == NULL)
			return NULL;

		scaled = pixbuf;
		pixbuf = gdk_pixbuf_add_alpha(pixbuf, FALSE, 0, 0, 0);
		gdk_pixbuf_unref(scaled);

		data = gdk_pixbuf_get_pixels(pixbuf);
		for (y = 0; y < gdk_pixbuf_get_height(pixbuf); y++)
			for (x = 0; x < gdk_pixbuf_get_width(pixbuf); x++) {
				data += 3;
				*data++ =
				    gdk_image_get_pixel(image, x,
							y) == 0 ? 0 : 255;
			}

		gdk_pixmap_unref(mask);
		gdk_image_destroy(image);
	}

	gdk_pixmap_unref(pixmap);

	XFree(wmhints);
	return pixbuf;
}

static gpointer
get_typed_property_data (Display *xdisplay,
			 Window   xwindow,
			 Atom     property,
			 Atom     requested_type,
			 gint    *size_p,
			 guint    expected_format)
{
	static const guint prop_buffer_lengh = 1024 * 1024;
	unsigned char *prop_data = NULL;
	Atom type_returned = 0;
	unsigned long nitems_return = 0, bytes_after_return = 0;
	int format_returned = 0;
	gpointer data = NULL;
	gboolean abort = FALSE;
	
	g_return_val_if_fail (size_p != NULL, NULL);
	*size_p = 0;
	
	gdk_error_trap_push ();
	
	abort = XGetWindowProperty (xdisplay,
				    xwindow,
				    property,
				    0, prop_buffer_lengh,
				    False,
				    requested_type,
				    &type_returned, &format_returned,
				    &nitems_return,
				    &bytes_after_return,
				    &prop_data) != Success;
	if (gdk_error_trap_pop () ||
	    type_returned == None)
		abort++;
	if (!abort &&
	    requested_type != AnyPropertyType &&
	    requested_type != type_returned) {
		g_warning (G_GNUC_PRETTY_FUNCTION "(): Property has wrong type, probably on crack");
		abort++;
	}
	if (!abort && bytes_after_return) {
			g_warning (G_GNUC_PRETTY_FUNCTION "(): Eeek, property has more than %u bytes, stored on harddisk?",
				   prop_buffer_lengh);
			abort++;
	}
	if (!abort && expected_format && expected_format != format_returned) {
		g_warning (G_GNUC_PRETTY_FUNCTION "(): Expected format (%u) unmatched (%d), programmer was drunk?",
			   expected_format, format_returned);
		abort++;
	}
	if (!abort && prop_data && nitems_return && format_returned) {
		switch (format_returned) {
		case 32:
			*size_p = nitems_return * 4;
			if (sizeof (gulong) == 8) {
				guint32 i, *mem = g_malloc0 (*size_p + 1);
				gulong *prop_longs = (gulong*) prop_data;
				
				for (i = 0; i < *size_p / 4; i++)
					mem[i] = prop_longs[i];
				data = mem;
			}
			break;
		case 16:
			*size_p = nitems_return * 2;
			break;
		case 8:
			*size_p = nitems_return;
			break;
		default:
			g_warning ("Unknown property data format with %d bits (extraterrestrial?)",
				   format_returned);
			break;
		}
		if (!data && *size_p) {
			guint8 *mem = g_malloc (*size_p + 1);
			
			memcpy (mem, prop_data, *size_p);
			mem[*size_p] = 0;
			data = mem;
		}
	}

	if (prop_data)
		XFree (prop_data);
	
	return data;
}


int on_menu(GtkMenuItem * menuitem, gpointer user_data)
{
	GwmhTask *task;
	GwmhDesk *desk_info;

	task = ((mytask *) user_data)->gwmhtask;
	desk_info = gwmh_desk_get_config();
	gwmh_desk_set_current_area(task->desktop, task->harea, task->varea);
	gwmh_task_show(task);
	gwmh_task_focus(task);
	return TRUE;
}

void init_task(GwmhTask * gwmh_task)
{
	mytask *task;
	GtkWidget *icon;
	GdkPixbuf *scaled;
	
	task = (mytask *) g_malloc0(sizeof(mytask));
	task->gwmhtask = gwmh_task;
	task->icon = get_icon(task, config->icon_size);
	tasklist = g_list_append(tasklist, task);
	task->item = gtk_pixmap_menu_item_new();
	task->label = gtk_label_new(gwmh_task->name);
	task->pixmap = get_pixmap(task);

	gtk_container_add(GTK_CONTAINER(task->item), task->label);

	gtk_pixmap_menu_item_set_pixmap(GTK_PIXMAP_MENU_ITEM(task->item),
					task->icon);
	gtk_widget_show(task->icon);

	if (!GWMH_TASK_SKIP_TASKBAR(task->gwmhtask)) {
		gtk_widget_show(task->label);
		gtk_widget_show(task->item);
	}

	gtk_signal_connect(GTK_OBJECT(task->item), "activate",
			   GTK_SIGNAL_FUNC(on_menu), task);

	gtk_menu_append(GTK_MENU(menu), task->item);

	if (GWMH_TASK_FOCUSED(task->gwmhtask)) {
		gtk_label_set_text(GTK_LABEL(label), task->gwmhtask->name);
		scaled = gdk_pixbuf_scale_simple(task->cache, 15, 15, 
				GDK_INTERP_HYPER);
		icon = get_pixmap_from_pixbuf(scaled);
		gdk_pixbuf_unref(scaled);
		if (GTK_BIN(pfrm)->child)
			gtk_widget_destroy(GTK_BIN(pfrm)->child);
		GTK_BIN(pfrm)->child = NULL;
		gtk_container_add(GTK_CONTAINER(pfrm), icon);
	}

	gtk_signal_connect(GTK_OBJECT(task->item), "button-press-event",
			GTK_SIGNAL_FUNC(on_item_click), task);
}

void update_task_icon(GwmhTask *gwmh_task) {
	mytask *task=find_task(gwmh_task);
	if (task->icon != NULL) {
		gtk_widget_destroy(task->icon);
		GTK_PIXMAP_MENU_ITEM(task->item)->pixmap = NULL;
	}

	if (task->cache != NULL) 
		gdk_pixbuf_unref(task->cache);

	task->icon = get_icon(task,config->icon_size);
	gtk_widget_show(task->icon);
	gtk_pixmap_menu_item_set_pixmap(
			GTK_PIXMAP_MENU_ITEM(task->item),task->icon);
}

void update_task(GwmhTask * gwmh_task)
{
	mytask *task;
	GtkWidget *icon;
	GdkPixbuf *scaled;

	task = find_task(gwmh_task);
	gtk_label_set_text(GTK_LABEL(task->label), task->gwmhtask->name);

	if (GWMH_TASK_FOCUSED(task->gwmhtask)) {
		gtk_label_set_text(GTK_LABEL(label), task->gwmhtask->name);
		if (GTK_BIN(pfrm)->child != NULL) {
			gtk_widget_destroy(GTK_BIN(pfrm)->child);
			GTK_BIN(pfrm)->child = NULL;
		}
		scaled = gdk_pixbuf_scale_simple(task->cache, 15, 15, 
				GDK_INTERP_HYPER);
		icon = get_pixmap_from_pixbuf(scaled);
		gdk_pixbuf_unref(scaled);
		gtk_widget_show(icon);
		gtk_container_add(GTK_CONTAINER(pfrm), icon);
	}
}

void destroy_task(GwmhTask * gwmh_task)
{
	mytask *task;

	task = find_task(gwmh_task);
	if (task == NULL)
		return;
	tasklist = g_list_remove(tasklist, task);
	if (task->icon)
		gtk_widget_destroy(task->icon);
	if (task->item != NULL)
		gtk_widget_destroy(task->item);
	if (task->cache)
		gdk_pixbuf_unref(task->cache);
	
	g_free(task);

}

mytask *find_task(GwmhTask * gwmh_task)
{
	GList *temp_tasks;
	mytask *task;

	temp_tasks = tasklist;

	while (temp_tasks) {
		task = (mytask *) temp_tasks->data;
		if (task->gwmhtask == gwmh_task)
			return task;
		temp_tasks = temp_tasks->next;
	}

	return NULL;
}

gint task_notifier(gpointer func_data, GwmhTask * gwmh_task,
		   GwmhTaskNotifyType ntype, GwmhTaskInfoMask imask)
{

	switch (ntype) {
	case GWMH_NOTIFY_NEW:
		init_task(gwmh_task);
		break;
	case GWMH_NOTIFY_DESTROY:
		destroy_task(gwmh_task);
		break;
	case GWMH_NOTIFY_INFO_CHANGED:
		if (imask & GWMH_TASK_INFO_WM_HINTS)
			if (find_task(gwmh_task)->pixmap != 
					get_pixmap(find_task(gwmh_task)))
				update_task_icon(gwmh_task);
		update_task(gwmh_task);
		break;
	default:
		break;
	}
	return TRUE;
}

void doabout(AppletWidget * widget, gpointer data)
{

	gchar *authors[] = {
		"mpalzcew@u.washington.edu",
		NULL
	};

	gtk_widget_show(gnome_about_new(PACKAGE, VERSION,
					"Copyright 2000",
					(const gchar **) authors,
					"MacOS like finder applet", NULL));

}

gint main(gint argc, gchar * argv[])
{
	/* I hate long functions */

	GtkWidget *tearoff;
	GtkWidget *hbox;
	gint i;

	/* setup pixmaps */
	i = strlen(PACKAGE_PIXMAPS_DIR);
	i += strlen("/unknown.png");
	i++;
	normal_path = g_malloc(i * sizeof(char));

	sprintf(normal_path, "%s%s", PACKAGE_PIXMAPS_DIR, "/unknown.png");

	/* init */
	applet_widget_init(PACKAGE, VERSION, argc, argv, NULL, 0, NULL);
	config = get_conf();

	/* new */
	applet = applet_widget_new(PACKAGE);
	tearoff = gtk_tearoff_menu_item_new();
	label = gtk_label_new("");
	menu = gtk_menu_new();
	hbox = gtk_hbox_new(FALSE, 0);
	pfrm = gtk_frame_new(NULL);
	frame = gtk_frame_new(NULL);

	/* pack */
	applet_widget_add(APPLET_WIDGET(applet), frame);
	gtk_container_add(GTK_CONTAINER(frame), hbox);
	gtk_box_pack_start(GTK_BOX(hbox), pfrm, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_menu_append(GTK_MENU(menu), tearoff);

	/* properties */
	gtk_widget_set_usize(applet, config->width,APPLET_WIDGET(applet)->size);
	gtk_widget_set_events(applet, GDK_BUTTON_PRESS_MASK);
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
	gtk_frame_set_shadow_type(GTK_FRAME(pfrm), GTK_SHADOW_NONE);
	gtk_frame_set_shadow_type(GTK_FRAME(frame), config->shadow_type);

	/* gtk_widget_show */
	gtk_widget_show(frame);
	gtk_widget_show(hbox);
	gtk_widget_show(pfrm);
	gtk_widget_show(tearoff);
	gtk_widget_show(menu);
	gtk_widget_show(label);
	gtk_widget_show(applet);

	/* gwmh */
	gwmh_init();
	gwmh_task_notifier_add(task_notifier, NULL);

	/* Callbacks */
	applet_widget_register_stock_callback((AppletWidget *) applet,
					      "Properties",
					      GNOME_STOCK_MENU_PROP,
					      "Properties", doproperties,
					      NULL);
	applet_widget_register_stock_callback((AppletWidget *) applet,
					      "About...",
					      GNOME_STOCK_MENU_ABOUT,
					      "About...", doabout, NULL);

	gtk_signal_connect(GTK_OBJECT(applet), "button-press-event",
			   GTK_SIGNAL_FUNC(on_click), NULL);
	gtk_signal_connect(GTK_OBJECT(applet), "change_pixel_size",
			GTK_SIGNAL_FUNC(on_size),NULL);

	/* cleanup and exit */
	gtk_main();
	free_conf(config);
	g_free(normal_path);
	return 0;
}
